#ifndef FORMULA_HH
#define FORMULA_HH

#include <iostream>
#include <cstdlib>
#include <cstring>
#include <iomanip>
#include <sstream>
#include <string>
#include <time.h>

#include "bdd.h"
#include "glpk.h"

#include <vector>
#include <utility>

#include "google/dense_hash_map"
#include <ext/hash_map>
/* will become #include <unordered_set> */

#include "rational.h"
#include "../rules/premise.h"
#include "../rules/setofconclusions.h"
#include "../satisfyingstack.h"

using namespace std;

using std::cout;
using std::endl;
using std::vector;
using google::dense_hash_map;
using __gnu_cxx::hash;

// Formula interface so we can have abstract template class pointers.
class IFormula {
	public:
		virtual bdd variable(int n)=0;
		virtual void set_bdd(bdd &b)=0;
		virtual bdd get_bdd()=0;
		virtual bdd modal(bdd *b, int n, int m)=0;
		virtual bool satisfiability(int option)=0;
		virtual void clear_maps()=0;
};

template<class ModalValueType>
class Formula : public IFormula{
	protected:
		/* REQUIRED DEFINITIONS FOR HASH_MAPS */		
		class data {
			public:
				ModalValueType value;
				bdd* argument;
				data() : value(-1) {argument = NULL; };
				data(ModalValueType v, bdd* arg) : value(v){ argument = arg; };
				~data(){ if(argument) delete argument; };
				data& operator=(const data& other){
					if (this!=&other){
						if(argument) 
							delete argument;
						value = other.value;
						argument = NULL;
						if(other.argument)
							argument = new bdd(*(other.argument));
						}
					return *this;
				};
				data(const data& other){
					value = other.value;
					argument = NULL;
					if(other.argument)
						argument = new bdd(*(other.argument));
				};
		};
	
		// Required for dense_hash_map - how to compare keys.
		struct eqkey {
			bool operator()(const string s1, const string s2) const{
				return s1 == s2;
			}

			bool operator()(const int s1, const int s2) const{
				return (s1 == s2);
			}

			bool operator()(const data& d1, const data& d2) const{
				return ((d1.value == d2.value) && ((d1.argument==d2.argument) || (d1.argument && d2.argument && (*(d1.argument) == *(d2.argument)))));
			}
		};		

		// Number of variables. Set to zero by constructor and
		// incremented as variables and modal variables are added.
		// (Used for building the hash_maps).	
		int number_of_variables;

		// Map of parsed variable numbers to BDD package numbers
		dense_hash_map<int, int, hash<int>, eqkey> variables;
		// Map of BDD package numbers to input numbers
		dense_hash_map<int, int, hash<int>, eqkey> variables_back;	
		
		// How to hash data items.
		struct datahash {
			hash <const char*> charhash;
			hash <int> inthash;
			size_t operator ()(const data& d) const {
				stringstream ss;
				string s;
				ss << d.value << " " << *d.argument << '\0';
				s = ss.str();
				char buf[s.size()+1];
				for(unsigned int i=0; i < s.size(); i++)
					buf[i] = s[i];
				buf[s.size()]='\0';
				return charhash(buf);
			}	
		};
		
		dense_hash_map<data, int, datahash, eqkey> modal_variables;
		dense_hash_map<int, data, hash<int>, eqkey> modal_variables_back;

		dense_hash_map<Premise<ModalValueType>, SetOfConclusions, typename Premise<ModalValueType>::Premise_hash, typename Premise<ModalValueType>::Premise_equal> rule_cache;

		int test_counter;

		bool check_satisfiability(bdd b);
		bool apply_rules(SatisfyingAssignment& sat);
		bool enumerate_rules(Premise<ModalValueType>& prem, bdd** underlying_formulae);
		bool enum_rules_intern(glp_prob* problem, glp_iocp* parameters, Premise<ModalValueType>& prem, bdd** underlying_formulae, vector<vector<bool> >& node_queue, vector<vector<bool> >& valid_nodes, vector<pair<vector<bool>, int> >& anti_nodes);
		bool run_solver(glp_prob *problem, glp_iocp* parameters, bool strict, vector<bool> current_node, int index);
		
		virtual void load_linear_program(glp_prob* problem, Premise<ModalValueType>& prem)=0;
	
	public:
		Formula();
		// The parsed formula
		bdd bdd_rep;

		// Used by the driver to set bdd_rep
		void set_bdd(bdd &b){ bdd_rep=b; };
		bdd get_bdd(){ return bdd_rep;};
		
		// Called by parser to construct variable or modal variable.
		// Also constructs appropriate data structures. 
		bdd variable(int n);
		bdd modal_var(bdd *b, ModalValueType &n);
		
		void print_back_vars();
		//virtual void print_back_modal()=0;
		
		bool satisfiability(int option){ return check_satisfiability(bdd_rep); };

		void clear_maps(){ 
			variables.clear();
			variables_back.clear();
			modal_variables.clear();
			modal_variables_back.clear();
			rule_cache.clear(); };
};
#endif
